/*	$Id: coda.c,v 1.29 2006-08-10 23:27:15 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/

///////////////////////////////////////////////////////////////////////////////
//
//  coda.c -- Coda File System Main Module
//
//  (C) Copyright 1996 OSR Open Systems Resources, Inc.
//  All rights reserved
//
//  OSR Open Systems Resources, Inc.
//  105 Route 101A Suite 19
//  Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//
///////////////////////////////////////////////////////////////////////////////

#define KERNEL 1

#include "coda_types.h"
#include "coda.h"
#include "coda_nt.h"
#include "ipc.h"
#include "cnode.h"
#include "debug.h"

//
// Common function declarations
//

VOID CodaDriverUnload(PDRIVER_OBJECT DriverObject);

// Temp debug/testing.  Need to move if kept!
NTSTATUS CodaIoCtrl(IN FS_FILE_HANDLE FileHandle, IN ULONG ControlCode,
		    IN PVOID InputBuffer, IN ULONG InputBufferSize,
		    IN PVOID OutputBuffer, IN ULONG OutputBufferSize,
		    OUT PULONG BytesWritten, OUT PULONG BytesNeeded);

//
// GLOBALS
//
int coda_mounted = 0;

struct coda_ntvolume coda_nt_volume;

// Global data for work queues
PDEVICE_OBJECT coda_device_object; 


#if DBG
//
// Debug -- Shared with CodaWRAP.C
//
CHAR LastFileName[CODA_MAXNAMLEN] = {0};
CHAR LastDirName[CODA_MAXNAMLEN] = {0};
extern FS_FILE_HANDLE LastDirectoryHandle= NULL;

#endif

// main file: here we instantiate data structures
// and register with the wrapper
static FS_OPERATIONS CodaOps = {0};
static HANDLE RegistrationHandle; 

//
// Basic forward declarations
//
NTSTATUS DriverEntry(PDRIVER_OBJECT, PUNICODE_STRING);

//
// DriverEntry
//
// This is the initialization entry point for the Driver.
//
// PARAMETERS:
//      DriverObj       Provided by the IoManager, describes THIS driver's
//                      instance
//      RegistryPath    Used as part of config, again provided by IoManager
//
//
// RETURNS:
//      NTSTATUS        result of DriverEntry. STATUS_SUCCESS indicates
//                      successful initialization.  All others result in
//                      the unloading of the driver.
//
//

NTSTATUS DriverEntry(PDRIVER_OBJECT DriverObj, PUNICODE_STRING RegistryPath)
{
    NTSTATUS  Status;
    PWCHAR    OSRCodaName = L"codadev";  // registered under Devices
    int       numberOfWorkThreads = 4;
    
    DbgPrint("Welcome to Coda! Version Beta-1.\n");
    DbgPrint("    Compiled on %s at %s.\n", __DATE__, __TIME__);
    
    DriverObj->DriverUnload = CodaDriverUnload;

    //
    // initialize rpc queues
    
    Status = request_init();
    inode_init();
    
    //
    // Init the Operations structure.  This is required.
    // ALL unused space must be initialized to zero.
    //
    RtlZeroMemory(&CodaOps, sizeof(FS_OPERATIONS));
    RtlZeroMemory(&coda_nt_volume, sizeof(coda_nt_volume));
    
    //
    // Fill in FLAGS long-word.  These are the standard
    // NT Volume Attributes, but applied to the file system
    // as a whole
    //
    CodaOps.Flags = (FILE_CASE_SENSITIVE_SEARCH
		     | FILE_CASE_PRESERVED_NAMES );

    CodaOps.Attributes = (
  //			  OW_FS_ATTR_NO_DIR_CACHE |
  //			  OW_FS_ATTR_NO_FILE_CACHE |
  			  OW_FS_ATTR_NO_OPLOCKS |
			  0);
    
    //
    // Fill in our operations table
    //
    // For the purposes of documentation:
    //  - Items not supported for network file systems do not appear below.
    //  - Items set to NULL below are supported by the wrapper for network
    //      file systems, but are not implemented for THIS network file 
    //      system at this time.
    //
    
    CodaOps.FsName.Buffer        = OSRCodaName;
    CodaOps.FsName.Length        = wcslen(OSRCodaName) * sizeof(WCHAR);
    CodaOps.FsName.MaximumLength = CodaOps.FsName.Length;
    
    CodaOps.FsCtrl               = coda_fsctrl;
    CodaOps.GetVolAttributes     = coda_get_volume_attributes;
    CodaOps.Mount                = coda_mount;
    CodaOps.Unmount		 = coda_unmount;
    CodaOps.Lookup               = coda_lookup;
    CodaOps.LookupById           = NULL;         // Not in this version
    CodaOps.SetAccessMode        = NULL;         // Not in this version
    CodaOps.Release              = coda_release;
    CodaOps.Create               = coda_create;
    CodaOps.Delete               = NULL;         // Using delete2
    CodaOps.Rename               = NULL;         // Using rename2
    CodaOps.Access               = coda_access;
    CodaOps.GetAttributes        = coda_getattr; 
    CodaOps.SetAttributes        = coda_setattr; 
    CodaOps.Flush                = NULL;         // Not in this version
    CodaOps.SetLength            = coda_set_length;
    CodaOps.Read                 = coda_read; 
    CodaOps.Write                = coda_write; 
    CodaOps.Ioctl                = coda_ioctl;
    CodaOps.ReadDirectory        = coda_readdir; // Convert to readdir2?
    
    CodaOps.QueryPath            = NULL;         // Not in this version
    CodaOps.Connect              = NULL;         // Not in this version
    CodaOps.Disconnect           = NULL;         // Not in this version
    CodaOps.UncRoot              = NULL;         // Not in this version
    CodaOps.Lock                 = NULL;         // Not in this version
    CodaOps.Unlock               = NULL;         // Not in this version

    CodaOps.UpdateShareAccess	 = coda_updshare; // Does nothing
    CodaOps.RemoveShareAccess    = coda_remshare; // Used like close.
    
    // FSDK v2 ones...
    CodaOps.Verify               = coda_verify;  // Required, but "NULL".
    
    CodaOps.Delete2              = coda_delete2;	
    CodaOps.Delete3              = NULL;	 // Needs conversion.
    CodaOps.ReadDirectory2       = NULL;
    CodaOps.ReadDirectory3       = NULL;	 // Needs conversion.
    CodaOps.Rename2              = coda_rename2;
    CodaOps.ReadCompressed       = NULL;	     // Not in coda
    CodaOps.WriteCompressed      = NULL;	     // Not in coda
    CodaOps.PagefileRead         = NULL;    	     // Not needed
    CodaOps.PagefileWrite        = NULL;	     // Not needed
    CodaOps.Clear                = NULL;	     // Not in coda
    CodaOps.GetNames             = NULL;             // Not in coda
    
    //
    // Register with the wrapper
    //
    Status = OwRegister(DriverObj, RegistryPath, OSR_WRAPPER_VERSION,
			OW_FS_TYPE_PSEUDO, &CodaOps, &RegistrationHandle);
    
    //
    // If we didn't register successfully, we're DONE
    //

    if ( !NT_SUCCESS(Status) ) {
	DbgPrint("DriverEntry: Problems in OwRegister: return code 0x%x\n",
		 Status);
	return(Status);
    }

#ifdef DBG
    // Initialize the user message list.
    init_msg_list();
#endif
    
    //
    // If the flow of control reaches this point, the call has been
    // successful.
    //
    
    CDEBUG((D_SUPER, "Coda: Leaving DriverEntry with SUCCESS.\n"));
    
    return (STATUS_SUCCESS);
}

//
// CodaDriverUnload
//
VOID CodaDriverUnload(PDRIVER_OBJECT DriverObject)
{
    NTSTATUS Status;
    
    CDEBUG((D_SUPER, "CodaDriverUnload: Unloading Driver...\n"));
    Status = OwDeregister(RegistrationHandle);
    
    if ( Status == STATUS_SUCCESS ) {
	CDEBUG((D_SUPER, "CodaDriverUnload: done\n"));
    } else {
	CDEBUG((D_SUPER, "CodaDriverUnload: "
	       "error encountered (exit %d)\n", Status));
    }
    
}


//
// CodaFsctrl
//
// This function lives in this module mainly for historical reasons.
// In future versions of the Coda driver, it should be moved to CodaWRAP.C
//
NTSTATUS coda_fsctrl(IN FS_FILE_HANDLE FileHandle, IN ULONG ControlCode,
		    IN PVOID InputBuffer, IN ULONG InputBufferSize,
		    IN PVOID OutputBuffer, IN ULONG OutputBufferSize,
		    OUT PULONG BytesWritten, OUT PULONG BytesNeeded)
{
    NTSTATUS Status = STATUS_SUCCESS;
    static FS_VOL_HANDLE FsdVolumeHandle = NULL;
    static FS_FILE_HANDLE DirectoryHandle= NULL;

#ifdef DBG    
    if ( ControlCode != CODA_FSCTL_GETDBGMSG)  
	CDEBUG((D_IPC | D_DETAILS, "coda_fsctrl: Starting with code: %x\n",
		ControlCode));		
#endif
    
    switch (ControlCode) {
	
#ifdef DBG
	//
	// Echo the input buffer back out
	//
    case CODA_FSCTL_ECHO:
	{
	    CDEBUG((D_IPC,
		   "coda_fsctrl: Doing FSCTL ECHO, "
		   "size: %d, in: %s",
		   InputBufferSize, InputBuffer));
	    if ( OutputBufferSize >= InputBufferSize ) {
		RtlCopyMemory( OutputBuffer, InputBuffer,
			       InputBufferSize);
		*BytesWritten = InputBufferSize;
	    } else {
		RtlCopyMemory(OutputBuffer, InputBuffer,
			      OutputBufferSize);
		*BytesWritten = OutputBufferSize;
	    }
	    
	    break;
	}
#endif
	
    case CODA_FSCTL_ANSWER:
	///	
	/// ANSWER_REQUEST -- venus -> kernel communication.
	///    Both upcall replies and downcalls.
	///
	{
	    struct request *req;
	    union outputArgs *out;
	    DWORD answerlen;
	    u_long unique;
	    u_long opcode;
	    unsigned long opcodebuf[2];
	    
	    /* Peek at the opcode, unique id */
	    
	    CDEBUG((D_IPC, "coda_fsctrl: ANSWER received %d bytes\n",
		   InputBufferSize));
	    if ( InputBufferSize > 2 * sizeof(u_long)) {
		RtlCopyMemory(opcodebuf, InputBuffer,
			      2 * sizeof(u_long));
	    } else {
		CDEBUG((D_IPC,
		       "coda_fsctrl: Not enough to "
		       "interpret answer\n!"));
		return STATUS_INVALID_DEVICE_REQUEST;
	    }
	    
	    opcode = opcodebuf[0];
	    
	    // *** at this point handle downcalls
	    
	    if (DOWNCALL(opcode)) {
		CDEBUG((D_IPC, "coda_fsctrl: ANSWER is downcall\n"));
		out = (union outputArgs *) InputBuffer;
		Status = coda_downcall (opcode, out);
		break;
	    }
	    
	    // request_find dequeues
	    unique = opcodebuf[1];
	    CDEBUG((D_IPC, "coda_fsctrl: ANSWER to unique=%d\n", unique));
	    req = request_find(unique);
	    if ( !req ) {
		CDEBUG((D_IPC, "coda_fsctrl: request lost!\n"));
		*BytesWritten = -1;
		break;
	    }
	    
	    /* move data into response buffer. */
	    out = req->req_data.out;
	    answerlen = MIN(InputBufferSize, req->req_bufflen);
	    RtlCopyMemory(out, InputBuffer, answerlen);
	    req->req_answerlen = answerlen;	
	    req->req_state = VM_WRITE;
	    if ( answerlen < InputBufferSize )
		DbgPrint("Answer bigger than bufflen for"
			 " opcode %ld, unique %ld",
			 opcode, unique);
	    
	    CDEBUG((D_IPC, "coda_fsctrl: Ready to signal caller\n"));
	    
	    // wake up caller
	    request_signal(req);
	    *BytesWritten = InputBufferSize;
	    CDEBUG((D_IPC, "coda_fsctrl: ANSWER done\n"));
	    break;
	}
	
    case CODA_FSCTL_FETCH:
	// 
	// Venus request for an upcall message.  request_get_pending()
	// times out so the venus thread is not continually blocked.
	// It appears that a process can not be called if a thread
	// is blocked in this call.
	// 
	//
	{
	    struct request *req;
	    
	    CDEBUG((D_IPC | D_DETAILS, "coda_fsctrl: Doing FETCH\n"));
	    
	    /* Get message at head of pending queue. */
	    req = request_get_pending(1);
	    if ( req ) {
		RtlCopyMemory(OutputBuffer, req->req_data.out,
			      req->req_fetchlen);
		*BytesWritten = req->req_fetchlen;
		req->req_state = VM_READ;
		
		CDEBUG((D_IPC,
		       "coda_fsctrl: FETCH: Got message with unique %d\n",
		       req->req_data.in->ih.unique));
	    } else {
		*BytesWritten = 0;
	    }
	    
	    break;
	}

#ifdef DBG	
    case CODA_FSCTL_TRYREQUEST:
	// a raw way to give the kernel Coda requests.  This is useful
	// for testing and can function as the pioctl call.
	{
	    long res;
	    struct request *req;
	    ULONG insize, outsize;
	    void *d;
	    
	    DbgPrint ("TRYREQUEST (%x).\n", CODA_FSCTL_TRYREQUEST);
	    req = request_allocate();
	    if (!req) {
		DbgPrint ("No memory for request struct.\n");
		break;
	    }
	    d = request_allocate_data(req, InputBufferSize);
	    if  ( !d  ) {
		DbgPrint("No memory for request data.\n");
		break;
	    }
	    RtlCopyMemory(req->req_data.in, InputBuffer,
			  InputBufferSize);
	    req->req_fetchlen = InputBufferSize;
	    
	    DbgPrint ("About to process_rpc\n");
	    res = request_process(req);
	    
	    DbgPrint ("process_rpc returned %d\n", res);
	    outsize = MIN(req->req_bufflen, OutputBufferSize);
	    RtlCopyMemory(OutputBuffer, req->req_data.out,
			  outsize);
	    *BytesWritten = outsize;
	    request_free_data(req);
	    request_free (req);
	    break;
	}
	
    case CODA_FSCTL_TRYUPCALL:
	// a raw way to give the kernel Coda requests.  This is useful
	// for testing and can function as the pioctl call.
	{
	    long res;
	    ULONG insize, outsize;
	    void *d;
	    DbgPrint("TRYUPCALL (%x).\n", CODA_FSCTL_TRYUPCALL);
	    insize = InputBufferSize;
	    outsize = InputBufferSize;
	    
	    d = coda_malloc(InputBufferSize);
	    if  ( !d  ) {
		DbgPrint("No memory for request data.\n");
		break;
	    }
	    RtlCopyMemory(d, InputBuffer, InputBufferSize);
	    
	    DbgPrint("making upcall: size %d\n", insize);
	    res = coda_upcall(insize, &outsize, d);
	    
	    DbgPrint ("coda_upcall returned %d\n", res);
	    if ( res == STATUS_SUCCESS) {
		outsize = MIN(outsize, OutputBufferSize);
		RtlCopyMemory(OutputBuffer, d, outsize);
	    } else {
		DbgPrint("Venus gives error: %d\n",res);
		outsize = 0;
	    }
	    *BytesWritten = outsize;
	    coda_free(d);
	    DbgPrint("UPCALL done\n");
	    break;
	}

    case CODA_FSCTL_SETDBGLEVEL: 
	/// 
	/// Set the debug level dynamically.
	///
	{   int uid;

	    if (InputBufferSize >= sizeof(ULONG)) {
		RtlCopyMemory (&DebugLevel, InputBuffer, sizeof(ULONG));
	    }
	    *BytesWritten = 0;
	    CDEBUG ((0, "coda_fsctrl: Set DebugLevel to 0x%x\n",
		     DebugLevel));

	    if (DebugLevel & D_PRCACHE)
		print_inode_cache();

	    break;
	}

    case CODA_FSCTL_GETDBGMSG: 
	/// 
	/// Get a debug message stored in the dgb message queue.
	///
	{
	    *BytesWritten = get_debug_mesg(OutputBuffer, OutputBufferSize);

	    // Get out without generating another debug message !
	    return STATUS_SUCCESS;  
	}
#endif

    case CODA_FSCTL_PIOCTL: 
	/// 
	/// Pioctl interface.  Done to "device file" rather
	/// than a weird file name in the root dir.
	///
	{
	    ULONG cmd;
	    void *buf;
	    ULONG bufsize;
	    ULONG outsize = OutputBufferSize;

	    bufsize = MAX(InputBufferSize, OutputBufferSize);

	    Status = STATUS_INVALID_DEVICE_REQUEST;
	    buf = coda_malloc(bufsize);
	    if  ( !buf  ) {
		DbgPrint("No memory for request data.\n");
		break;
	    }

	    RtlCopyMemory(buf, InputBuffer, InputBufferSize);

	    // Do the pioctl!
	    Status = coda_pioctl ((struct PioctlData*) buf, InputBufferSize,
				   OutputBuffer, &outsize);

	    if ( Status ) {
		DbgPrint("coda_fsctrl: coda_pioctl returns error %d\n",
			 Status);
	    }
	    *BytesWritten = outsize;
	    coda_free (buf);
	}
	break;

    default:
	DbgPrint("coda_fsctrl: Invalid request %d\n", ControlCode);
	Status = STATUS_INVALID_DEVICE_REQUEST;

    }
    
    CDEBUG((D_IPC | D_DETAILS, "coda_fsctrl: Returning %d\n", Status));
    return(Status);
}
